
/* 
  Copyright (C) 1994, 2002  Free Software Foundation, Inc.
  Copyright (C) 2005, 2008  Dmitry V. Levin <ldv@altlinux.org>

  front-compress encoding.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
*/

/* Uses front compression (also known as incremental encoding);
   see ";login:", March 1983, p. 8.

   The frcode() function gets NUL-terminated string and outputs encoded entity.
   Each entity consists of an offset-differential count byte (the additional
   number of characters of prefix of the preceding entry to use beyond the
   number that the preceding entry is using of its predecessor), followed by a
   null-terminated ASCII remainder.

   If the offset-differential count is larger than can be stored in a byte
   (+/-127), the byte has the value 0x80 and the count follows in a 2-byte
   word, with the high byte first (network byte order).

   Example:

   Input, with NULs changed to newlines:
   /usr/src
   /usr/src/cmd/aardvark.c
   /usr/src/cmd/armadillo.c
   /usr/tmp/zoo

   Length of the longest prefix of the preceding entry to share:
   0 /usr/src
   8 /cmd/aardvark.c
   14 rmadillo.c
   5 tmp/zoo

   Output, with NULs changed to newlines and count bytes made printable:
   0 LOCATE02
   0 /usr/src
   8 /cmd/aardvark.c
   6 rmadillo.c
   -9 tmp/zoo

   (6 = 14 - 8, and -9 = 5 - 14)

   Written by James A. Woods <jwoods@adobe.com>.
   Modified by David MacKenzie <djm@gnu.ai.mit.edu>.
   Modified by Dmitry V. Levin <ldv@altlinux.org>.
*/

#include <stdio.h>
#include <errno.h>
#include <error.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>

#include "locatedb.h"
#include "xmalloc.h"
#include "frencode.h"

#define xfwrite(ptr, size, fp) \
	do { \
		size_t n__ = (size); \
		if (fwrite((ptr), 1, n__, (fp)) != n__) \
			error(EXIT_FAILURE, errno, __func__); \
	} while (0)

#define xputc(c, fp) \
	do { \
		if (putc((c), (fp)) == EOF) \
			error(EXIT_FAILURE, errno, __func__); \
	} while (0)

/*
 * Return the length of the longest common prefix of strings S1 and S2.
 */

static inline size_t
prefix_length(const char *s1, const char *s2)
{
	register const char *start;

	for (start = s1; *s1 == *s2 && *s1 != '\0'; ++s1, ++s2)
		;
	return (size_t)(s1 - start);
}

void
frencode(FILE *fp, const char *path)
{
	static char *old_path;
	static size_t old_prefix_len;

	size_t  prefix_len = prefix_length(old_path ? : "", path);
	int     diff = (int)prefix_len - (int)old_prefix_len;
	size_t	path_len = strlen(path);

	old_path = xrealloc(old_path, path_len + 1);
	strcpy(old_path + prefix_len, path + prefix_len);
	old_prefix_len = prefix_len;

	static int initialized = 0;
	if (!initialized)
	{
		xfwrite(LOCATEDB_MAGIC, sizeof(LOCATEDB_MAGIC), fp);
		initialized = 1;
	}

	/* If the difference is small, it fits in one byte. */
	if ((diff >= -127) && (diff <= 127))
	{
		char    ch = (char) diff;

		xputc(ch, fp);
	} else
	/* Otherwise, two bytes plus a marker noting that fact.  */
	{
		/* XXX diff may overflow short int */
		char    buf[3] = { (char) 0x80, (char) (diff >> 8), (char) diff };

		xfwrite(buf, sizeof(buf), fp);
	}

	xfwrite(path + prefix_len, path_len - prefix_len + 1, fp);
}
